"""
Copyright 2021-2024 MICRORISC s.r.o.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import re
import time
from logging import Logger
from typing import List, Union

from packaging import version

from Setter.BootFlagManager import BootFlagManager
from Setter.Common.Command import Command
from Setter.Common.DaemonApi import DaemonApi
from Setter.Common.FileManager import FileManager
from Setter.Common.GatewayFile import GatewayFile
from Setter.Common.GatewayImage import GatewayImage
from Setter.Common.ServiceManager import ServiceManager
from Setter.DaemonConfig import DaemonConfig
from Setter.Enums.Software import SoftwareBackup, SystemdServices
from Setter.Exceptions.CommandException import CommandException
from Setter.Exceptions.IOErrorException import IOErrorException
from Setter.Exceptions.ServiceErrorException import ServiceErrorException
from Setter.IqarosV1Config import IqarosV1Config
from Setter.IqarosV2Config import IqarosV2Config
from Setter.IqrfCloudConfig import IqrfCloudConfig
from Setter.Logging import Logging
from Setter.MosquittoConfig import MosquittoConfig
from Setter.NetworkConfig import NetworkConfig
from Setter.WebsocketClient import WebsocketClient


class GatewayConfig:
	"""
	IQRF Gateway device configuration class.

	Methods
	-------
	configure(config: dict) -> None:
		Configure gateway for product utilizing gateway information file.
	get_image_version(image_version: str) -> None:
		Retrieve and store image version for later use.
	generate_gwid() -> None:
		Generate gateway id from mac address and module id.
	configure_tr() -> None:
		Configure module TR configuration.
	configure_system() -> None:
		Configure system time, hostname etc.
	generate_certificate() -> None:
		Generate certificate for daemon and webapp tls.
	backup_configuration() -> None:
		Backup configuration of gateway software and network.
	restart_services() -> None:
		Restart systemd services.
	check_boot_flag() -> bool:
		Check if flag preventing repeated configuration is present.
	check_image_version() -> None:
		Check and update image version in gateway file.
	"""

	# pylint: disable=unsubscriptable-object
	_id: str = None
	_host: str = None
	_backup_dir: str = '/usr/share/iqrf-gateway/default'
	_boot_flag: Union[BootFlagManager, None] = None
	_services: List[str] = [
		SystemdServices.AVAHI_DAEMON,
		SystemdServices.RESOLVCONF,
		SystemdServices.NETWORK_MANAGER,
		SystemdServices.MOSQUITTO,
		SystemdServices.DAEMON,
		SystemdServices.NGINX,
		SystemdServices.MENDER_CLIENT,
		SystemdServices.MENDER_CONNECT,
		SystemdServices.MODEM_MANAGER
	]
	_logger: Logger = Logging.get_logger()
	_image: Union[GatewayImage, None] = None

	@classmethod
	def configure(cls, config: dict) -> None:
		"""
		Configure gateway for product utilizing gateway information file.

		Parameters
		----------
		config: dict
			gateway information file

		Returns
		-------
		None
		"""
		cls._logger.info('Configuring IQRF Gateway device.')
		WebsocketClient.create_connection('localhost', 1338)

		cls.generate_gwid()
		config['gwId'] = cls._id.upper()
		config['gwHost'] = config['gwHost'].replace('ffffffffffffffff', cls._id)
		cls._host = str(config['gwHost'])
		if cls._image is not None:
			threshold = 'v1.2.0' if cls._image.product == 'iqaros' else 'v1.5.0'
			if cls._image.os_version >= version.parse(threshold) and 'gwToken' in config:
				config.pop('gwToken')

		MosquittoConfig.create_user('iqrf', cls._id.upper())
		DaemonConfig.configure_splitter(cls._id)
		DaemonConfig.configure_ide(config['gwProduct'])
		DaemonConfig.configure_mqtt(cls._host[-22:-6])
		GatewayFile.write(config)
		cls.configure_tr()
		cls.configure_system()
		cls.generate_certificate()

		if cls._image.product == 'iqaros':
			cls.configure_iqaros()

		IqrfCloudConfig.configure()

		WebsocketClient.close_connection()

		cls.backup_configuration()
		cls.restart_services()
		admin_username: str = 'admin' if cls._image.os_type == 'yocto' else 'root'
		cls.change_account_password(admin_username, cls._id.upper())
		cls._boot_flag.check_flag()

	@classmethod
	def configure_iqaros(cls) -> None:
		"""
		Configure IQAROS specific services.

		Returns
		------−
		None
		"""
		if cls._image.os_version.major == 1:
			IqarosV1Config.configure()
			cls._services += IqarosV1Config.get_services()
			DaemonConfig.configure_mqtt_iqaros(cls._host[-22:-6])
		elif cls._image.os_version.major == 2:
			IqarosV2Config.configure()
			cls._services += IqarosV2Config.get_services()
		else:
			cls._logger.error('Unsupported IQAROS version %s.', str(cls._image.os_version))

	@classmethod
	def get_image_version(cls, image_version: str) -> None:
		"""
		Retrieve and store image version for later use.

		Parameters
		----------
		image_version: str
			Gateway image version string.

		Returns
		-------
		None
		"""
		try:
			cls._image = GatewayImage.from_str(image_version)
		except ValueError:
			cls._logger.warning('Failed to match and store image version.')
		cls._boot_flag = BootFlagManager(cls._image)

	@classmethod
	def generate_gwid(cls) -> None:
		"""
		Generate gateway ID from mac address and module ID.

		Returns
		-------
		None
		"""
		mac = NetworkConfig.get_mac('eth0').replace(':', '')
		rsp = DaemonApi.read_os(0)
		mid = rsp['data']['rsp']['result']['mid']
		cls._id = mac[:6] + format(mid & 0xffff, '04x') + mac[6:]
		cls._logger.info('Generated GW ID: %s', cls._id)

	@classmethod
	def configure_tr(cls) -> None:
		"""
		Configure module TR configuration.

		Sets TX power and RX filter.

		Returns
		-------
		None
		"""
		rsp = DaemonApi.write_tr_config(0, {'txPower': 7, 'rxFilter': 7})
		if rsp['data']['rsp']['restartNeeded']:
			DaemonApi.restart_os(0)

	@classmethod
	def configure_system(cls) -> None:
		"""
		Configure system time, hostname etc.

		Returns
		-------
		None

		Exceptions
		----------
		CommandException
			raised if command to set timezone failed
		"""
		cls._logger.info('Configuring timezone Europe/Prague.')
		result = Command.execute('timedatectl set-timezone Europe/Prague', True)
		if result.returncode != 0:
			error_str = f'Failed to set timezone: {result.stderr}.'
			cls._logger.error(error_str)
			raise CommandException(error_str)

	@classmethod
	def generate_certificate(cls) -> None:
		"""
		Generate TLS certificate for daemon and webapp.

		Certificate directories are created if they do not exist.
		Existing certificates are removed and new certificates are generated.
		Newly generated files have their permissions fixed.

		Returns
		-------
		None

		Exceptions
		----------
		CommandException
			raised if certificate or private key command failed
		"""
		cls._logger.info('Generating certificates.')
		daemon_cert_path = '/etc/iqrf-gateway-daemon/certs/core/'
		webapp_cert_path = '/etc/iqrf-gateway-webapp/certs/'
		files = ['cert.pem', 'privkey.pem']

		if not os.path.exists(daemon_cert_path):
			os.makedirs(daemon_cert_path, exist_ok=True)
		if not os.path.exists(webapp_cert_path):
			os.makedirs(webapp_cert_path, exist_ok=True)

		# remove old files if they exist
		for file in files:
			if os.path.exists(webapp_cert_path + file):
				os.unlink(webapp_cert_path + file)
			if os.path.exists(daemon_cert_path + file):
				os.remove(daemon_cert_path + file)

		# generate private key
		command = f'openssl ecparam -name secp384r1 -genkey -param_enc named_curve -out {daemon_cert_path}{files[1]}'
		result = Command.execute(command, True)
		if result.returncode != 0:
			error_str = f'Failed to generate private key: {result.stderr}.'
			cls._logger.error(error_str)
			raise CommandException(error_str)

		# generate certificate
		command = 'openssl req -new -x509 -sha256 -nodes -days 3650 ' \
			'-subj "/CN=IQRF Gateway/C=CZ/ST=Hradec Kralove Region/L=Jicin/O=IQRF Tech s.r.o." ' \
			'-key {dir}{keyfile} -out {dir}{certfile}'.format(dir=daemon_cert_path, keyfile=files[1], certfile=files[0])
		result = Command.execute(command, True)
		if result.returncode != 0:
			error_str = f'Failed to generate certificate: {result.stderr}.'
			cls._logger.error(error_str)
			raise CommandException(error_str)

		# add symlinks to webapp
		for file in files:
			os.chmod(daemon_cert_path + file, 0o600)
			os.symlink(daemon_cert_path + file, webapp_cert_path + file)

	@classmethod
	def backup_configuration(cls) -> None:
		"""
		Backup configuration of gateway software and network.

		Returns
		-------
		None

		Raises
		------
		IOErrorException
			raised if an error occurs while copying configuration files for backup
		"""
		cls._logger.info('Backing up default configuration.')
		backup_sw = [SoftwareBackup.CONTROLLER, SoftwareBackup.DAEMON, SoftwareBackup.UPLOADER]

		# check for optional services
		if ServiceManager.exists(SystemdServices.CLOUD_PROVISIONING):
			backup_sw.append(SoftwareBackup.CLOUD_PROVISIONING)
		if ServiceManager.exists(SystemdServices.INFLUXDB_BRIDGE):
			backup_sw.append(SoftwareBackup.INFLUXDB_BRIDGE)
		if ServiceManager.exists(SystemdServices.NETWORK_MANAGER):
			backup_sw.append(SoftwareBackup.NETWORK_MANAGER)

		# generate backup directory tree
		if not os.path.isdir(cls._backup_dir):
			os.makedirs(cls._backup_dir)
		for software in backup_sw:
			directory = f'{cls._backup_dir}/{software}'
			if not os.path.isdir(directory):
				os.makedirs(directory)

		# backup files
		try:
			# backup gateway file
			FileManager.copy(
				source='/etc/iqrf-gateway.json',
				destination=f'{cls._backup_dir}/iqrf-gateway.json',
			)
			# backup controller
			FileManager.copy_tree(
				source='/etc/iqrf-gateway-controller/',
				destination=f'{cls._backup_dir}/{SoftwareBackup.CONTROLLER}',
			)
			# backup daemon
			FileManager.copy_tree(
				source='/etc/iqrf-gateway-daemon',
				destination=f'{cls._backup_dir}/{SoftwareBackup.DAEMON}/config',
			)
			FileManager.copy_tree(
				source='/usr/share/iqrf-gateway-daemon/cfgSchemas',
				destination=f'{cls._backup_dir}/{SoftwareBackup.DAEMON}/cfgSchemas',
			)
			FileManager.copy_tree(
				source='/usr/share/iqrf-gateway-daemon/schedulerSchemas',
				destination=f'{cls._backup_dir}/{SoftwareBackup.DAEMON}/schedulerSchemas',
			)
			FileManager.copy_tree(
				source='/var/cache/iqrf-gateway-daemon/scheduler',
				destination=f'{cls._backup_dir}/{SoftwareBackup.DAEMON}/scheduler',
			)
			# backup uploader
			FileManager.copy_tree(
				source='/etc/iqrf-gateway-uploader/',
				destination=f'{cls._backup_dir}/{SoftwareBackup.UPLOADER}',
			)
			# backup cloud provisioning
			if SoftwareBackup.CLOUD_PROVISIONING in backup_sw:
				FileManager.copy_tree(
					source='/etc/iqrf-cloud-provisioning/',
					destination=f'{cls._backup_dir}/{SoftwareBackup.CLOUD_PROVISIONING}/config',
				)
				FileManager.copy_tree(
					source='/usr/share/iqrf-cloud-provisioning/cert/',
					destination=f'{cls._backup_dir}/{SoftwareBackup.CLOUD_PROVISIONING}/certs',
				)
			# backup influxdb-bridge
			if SoftwareBackup.INFLUXDB_BRIDGE in backup_sw:
				FileManager.copy_tree(
					source='/etc/iqrf-gateway-influxdb-bridge/',
					destination=f'{cls._backup_dir}/{SoftwareBackup.INFLUXDB_BRIDGE}',
				)
			# backup network manager
			if SoftwareBackup.NETWORK_MANAGER in backup_sw:
				FileManager.copy(
					source='/etc/NetworkManager/NetworkManager.conf',
					destination=f'{cls._backup_dir}/{SoftwareBackup.NETWORK_MANAGER}/NetworkManager.conf',
				)
				FileManager.copy_tree(
					source='/etc/NetworkManager/system-connections',
					destination=f'{cls._backup_dir}/{SoftwareBackup.NETWORK_MANAGER}/system-connections',
				)
		except IOErrorException as io_error:
			cls._logger.error(str(io_error))
			raise io_error

	@classmethod
	def restart_services(cls) -> None:
		"""
		Restart systemd services.

		Returns
		-------
		None
		"""
		cls._logger.info('Restarting systemd services.')
		for service in cls._services:
			if not ServiceManager.exists(service):
				cls._logger.info('Service "%s" does not exist. Skipping...', service)
				continue
			try:
				if ServiceManager.is_running(service):
					cls._logger.info('Restarting %s service...', service)
					ServiceManager.restart(service)
			except ServiceErrorException as service_error:
				cls._logger.error('Failed to restart %s service. Error: %s', service, service_error.message)
			time.sleep(1)

	@classmethod
	def check_boot_flag(cls) -> bool:
		"""
		Check if flag preventing repeated configuration is present.

		Presence of flag file is checked before gateway configuration is executed.
		Gateway is configured if flag file does not exist or if utility is invoked with --force options.

		Returns
		-------
		bool
			True if flag file is present, False otherwise
		"""
		return cls._boot_flag.check_flag()

	@classmethod
	def check_image_version(cls, config: dict) -> None:
		"""
		Check and update image version in gateway file.

		Compares image version in system information and gateway file.
		If the versions mismatch, gateway file is updated according to system information.

		Parameters
		----------
		config: dict
			gateway information file

		Returns
		-------
		None

		Raises
		------
		IOErrorException
			raised if an error occurs while reading the system version
		"""
		if cls._image is None or cls._image.os_type != 'yocto':
			return
		try:
			found = False
			with open('/etc/os-release', mode='r', encoding='utf-8') as file:
				for line in file:
					if not re.search('^VERSION_ID=.*$', line):
						continue
					found = True
					system_version = version.parse('v' + line.split('=')[1].strip().replace('"', ''))
					break
			if not found:
				cls._logger.warning('Could not find OS version entry.')
				return
			if cls._image.os_version == system_version:
				return
			cls._logger.info('New image version detected, updating gateway file to %s.', system_version)
			cls._image.os_version = system_version
			config['gwImage'] = str(cls._image)
			GatewayFile.write(config)
		except IOError as io_error:
			error_str = Logging.io_error(io_error)
			cls._logger.error(error_str)
			raise IOErrorException(error_str) from io_error

	@classmethod
	def change_account_password(cls, username: str, password: str) -> None:
		"""
		Change password for specified account.

		Parameters
		----------
		username: str
			account name
		password: str
			new password

		Returns
		-------
		None

		Raises
		------
		CommandException
			raised if command to change password fails
		"""
		cls._logger.info('Changing password for account %s.', username)
		username = username.replace('\'', '\'"\'"\'')
		password = password.replace('\'', '\'"\'"\'')
		result = Command.execute(f'echo \'{username}:{password}\' | chpasswd', True)
		if result.returncode != 0:
			error_str = f'Failed to change password for account {username}: {result.stderr}.'
			cls._logger.error(error_str)
			raise CommandException(error_str)
